@@ -371,6 +371,17 @@ func TestFindDeepestTags(t *testing.T) {
 	assert.Equal("bar", tags[1])
 	assert.Equal(2, len(tags))
 
+	v4, _, _ = patricia.ParseIPFromString("1.2.3.5")
+	found, tags = tree.FindDeepestTagsWithFilter(*v4, func(t GeneratedType) bool { return t.(string) == "foo" })
+	assert.True(found)
+	assert.Equal("foo", tags[0])
+	assert.Equal(1, len(tags))
+
+	v4, _, _ = patricia.ParseIPFromString("1.2.3.5")
+	found, tags = tree.FindDeepestTagsWithFilter(*v4, func(t GeneratedType) bool { return t.(string) == "nothing" })
+	assert.True(found)
+	assert.Equal(0, len(tags))
+
 	// test searching for an address that has nothing
 	v4, _, _ = patricia.ParseIPFromString("9.9.9.9")
 	found, tags = tree.FindDeepestTags(*v4)
@@ -770,7 +781,7 @@ func TestSimpleTree(t *testing.T) {
 	// find deepest tag: no match
 	found, tag = tree.FindDeepestTag(ipv4FromBytes([]byte{128, 4, 3, 2}, 32))
 	assert.False(t, found)
-	assert.Nil(t, tag)
+	assert.Zero(t, tag)
 
 	// Add a couple root tags
 	countIncreased, count := tree.Add(ipv4FromBytes([]byte{127, 0, 0, 1}, 0), "root1", nil)
@@ -997,6 +1008,127 @@ func TestSet(t *testing.T) {
 	assert.Equal(t, "parent", tag)
 }
 
+func TestSetOrUpdate(t *testing.T) {
+	buf := make([]GeneratedType, 0)
+
+	address := ipv4FromBytes([]byte{1, 2, 3, 4}, 32)
+
+	tree := NewTreeV4()
+
+	// add a parent node, just to mix things up
+	countIncreased, count := tree.Set(ipv4FromBytes([]byte{1, 2, 3, 0}, 24), "parent")
+	assert.True(t, countIncreased)
+	assert.Equal(t, 1, count)
+
+	countIncreased, count = tree.SetOrUpdate(address, "tagA",
+		func(GeneratedType) GeneratedType { return "tagZ" })
+	assert.True(t, countIncreased)
+	assert.Equal(t, 1, count)
+	found, tag := tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagA", tag)
+
+	countIncreased, count = tree.SetOrUpdate(address, "tagZ",
+		func(old GeneratedType) GeneratedType { return old.(string) + "B" })
+	assert.Equal(t, 1, count)
+	assert.False(t, countIncreased)
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagAB", tag)
+
+	countIncreased, count = tree.SetOrUpdate(address, "tagZ",
+		func(old GeneratedType) GeneratedType { return old.(string) + "C" })
+	assert.Equal(t, 1, count)
+	assert.False(t, countIncreased)
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagABC", tag)
+
+	countIncreased, count = tree.SetOrUpdate(address, "tagZ",
+		func(old GeneratedType) GeneratedType { return old.(string) + "D" })
+	assert.Equal(t, 1, count)
+	assert.False(t, countIncreased)
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagABCD", tag)
+
+	// now delete the tag
+	delCount := tree.DeleteWithBuffer(buf, address, func(a GeneratedType, b GeneratedType) bool { return true }, "")
+	assert.Equal(t, 1, delCount)
+
+	// verify it's gone - should get the parent
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "parent", tag)
+}
+
+func TestAddOrUpdate(t *testing.T) {
+	buf := make([]GeneratedType, 0)
+
+	address := ipv4FromBytes([]byte{1, 2, 3, 4}, 32)
+
+	tree := NewTreeV4()
+
+	// add a parent node, just to mix things up
+	countIncreased, count := tree.Set(ipv4FromBytes([]byte{1, 2, 3, 0}, 24), "parent")
+	assert.True(t, countIncreased)
+	assert.Equal(t, 1, count)
+
+	countIncreased, count = tree.AddOrUpdate(address, "tagA",
+		nil,
+		func(GeneratedType) GeneratedType { return "tagZ" })
+	assert.True(t, countIncreased)
+	assert.Equal(t, 1, count)
+	found, tag := tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagA", tag)
+
+	countIncreased, count = tree.AddOrUpdate(address, "tagZ",
+		func(GeneratedType, GeneratedType) bool { return true },
+		func(old GeneratedType) GeneratedType { return old.(string) + "B" })
+	assert.Equal(t, 1, count)
+	assert.False(t, countIncreased)
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagAB", tag)
+
+	countIncreased, count = tree.AddOrUpdate(address, "tagAB",
+		func(val1 GeneratedType, val2 GeneratedType) bool { return val1 == val2 },
+		func(old GeneratedType) GeneratedType { return old.(string) + "C" })
+	assert.Equal(t, 1, count)
+	assert.False(t, countIncreased)
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "tagABC", tag)
+
+	countIncreased, count = tree.AddOrUpdate(address, "tagABCD",
+		func(val1 GeneratedType, val2 GeneratedType) bool { return val1 == val2 },
+		func(old GeneratedType) GeneratedType { return old.(string) + "Z" })
+	assert.Equal(t, 2, count)
+	assert.True(t, countIncreased)
+	found, tags := tree.FindDeepestTags(address)
+	assert.True(t, found)
+	assert.True(t, tagArraysEqual(tags, []string{"tagABC", "tagABCD"}))
+
+	countIncreased, count = tree.AddOrUpdate(address, "tagABC",
+		func(val1 GeneratedType, val2 GeneratedType) bool { return val1 == val2 },
+		func(old GeneratedType) GeneratedType { return old.(string) + "DE" })
+	assert.Equal(t, 2, count)
+	assert.False(t, countIncreased)
+	found, tags = tree.FindDeepestTags(address)
+	assert.True(t, found)
+	assert.True(t, tagArraysEqual(tags, []string{"tagABCDE", "tagABCD"}))
+
+	// now delete the tag
+	delCount := tree.DeleteWithBuffer(buf, address, func(a GeneratedType, b GeneratedType) bool { return true }, "")
+	assert.Equal(t, 2, delCount)
+
+	// verify it's gone - should get the parent
+	found, tag = tree.FindDeepestTag(address)
+	assert.True(t, found)
+	assert.Equal(t, "parent", tag)
+}
+
 func TestDelete1(t *testing.T) {
 	tags := make([]GeneratedType, 0)
 
@@ -1144,10 +1276,10 @@ func TestTagsMap(t *testing.T) {
 	tree := NewTreeV4()
 
 	// insert tags
-	tree.addTag("tagA", 1, nil, false)
-	tree.addTag("tagB", 1, nil, false)
-	tree.addTag("tagC", 1, nil, false)
-	tree.addTag("tagD", 0, nil, false) // there's no node0, but it exists, so use it for this test
+	tree.addTag("tagA", 1, nil, nil)
+	tree.addTag("tagB", 1, nil, nil)
+	tree.addTag("tagC", 1, nil, nil)
+	tree.addTag("tagD", 0, nil, nil) // there's no node0, but it exists, so use it for this test
 
 	// verify
 	assert.Equal(t, 3, tree.nodes[1].TagCount)
@@ -1259,3 +1391,225 @@ func TestDuplicateTagsWithMatchFunc(t *testing.T) {
 	assert.False(t, wasAdded)
 	assert.Equal(t, 2, count)
 }
+
+// test tree traversal
+func TestIterateV4(t *testing.T) {
+	tree := NewTreeV4()
+
+	// try an empty tree first
+	iter := tree.Iterate()
+	for iter.Next() {
+		assert.Fail(t, "empty tree should not have a next element")
+	}
+
+	ipA := ipv4FromBytes([]byte{203, 143, 220, 0}, 23)
+	ipB := ipv4FromBytes([]byte{203, 143, 220, 198}, 32)
+	ipC := ipv4FromBytes([]byte{203, 143, 0, 0}, 16)
+	ipD := ipv4FromBytes([]byte{203, 143, 221, 75}, 32)
+
+	// add the 4 addresses
+	tree.Add(ipA, "A", nil)
+	tree.Add(ipB, "B", nil)
+	tree.Add(ipC, "C", nil)
+	tree.Add(ipD, "D1", nil)
+	tree.Add(ipD, "D2", nil)
+
+	expected := map[string][]string{
+		"203.143.220.0/23":   {"A"},
+		"203.143.220.198/32": {"B"},
+		"203.143.0.0/16":     {"C"},
+		"203.143.221.75/32":  {"D1", "D2"},
+	}
+	got := map[string][]string{}
+	iter = tree.Iterate()
+	for iter.Next() {
+		tags := []string{}
+		for _, s := range iter.Tags() { //nolint:gosimple
+			tags = append(tags, s.(string))
+		}
+		got[iter.Address().String()] = tags
+	}
+	assert.Equal(t, expected, got)
+}
+
+// test deletion during tree traversal
+func TestIterateAndDeleteV4(t *testing.T) {
+	tree := NewTreeV4()
+
+	compare := func(expected [][]string) {
+		t.Helper()
+		got := [][]string{}
+		iter := tree.Iterate()
+		for iter.Next() {
+			tags := []string{}
+			for _, s := range iter.Tags() { //nolint:gosimple
+				tags = append(tags, s.(string))
+			}
+			got = append(got, append([]string{iter.Address().String()}, tags...))
+		}
+		assert.Equal(t, expected, got)
+	}
+
+	ipA := ipv4FromBytes([]byte{203, 143, 220, 0}, 23)
+	ipB := ipv4FromBytes([]byte{203, 143, 220, 198}, 31)
+	ipC := ipv4FromBytes([]byte{203, 143, 0, 0}, 16)
+	ipD := ipv4FromBytes([]byte{203, 143, 221, 75}, 32)
+	ipE := ipv4FromBytes([]byte{203, 143, 220, 198}, 32)
+
+	tree.Add(ipA, "A", nil)
+	tree.Add(ipB, "B", nil)
+	tree.Add(ipC, "C", nil)
+	tree.Add(ipD, "D1", nil)
+	tree.Add(ipD, "D2", nil)
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.0/23", "A"},
+		{"203.143.220.198/31", "B"},
+		{"203.143.221.75/32", "D1", "D2"},
+	})
+
+	// Delete one tag, no node
+	iter := tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "D1"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.0/23", "A"},
+		{"203.143.220.198/31", "B"},
+		{"203.143.221.75/32", "D2"},
+	})
+
+	// Delete a node with two children
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "A"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.198/31", "B"},
+		{"203.143.221.75/32", "D2"},
+	})
+
+	// Delete right children of a node two children
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "D2"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.198/31", "B"},
+	})
+
+	// Delete leaf
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "B"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+	})
+
+	// Delete root
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "C"
+		}, "")
+	}
+	compare([][]string{})
+
+	// Delete a node with a left children only
+	tree = NewTreeV4()
+	tree.Add(ipA, "A", nil)
+	tree.Add(ipB, "B", nil)
+	tree.Add(ipC, "C", nil)
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "A"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.198/31", "B"},
+	})
+
+	// Delete a node with a right children only
+	tree = NewTreeV4()
+	tree.Add(ipA, "A", nil)
+	tree.Add(ipC, "C", nil)
+	tree.Add(ipD, "D", nil)
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "A"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.221.75/32", "D"},
+	})
+
+	// Delete a node without children and at the left of its empty parent
+	tree = NewTreeV4()
+	tree.Add(ipB, "B", nil)
+	tree.Add(ipC, "C", nil)
+	tree.Add(ipD, "D", nil)
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "B"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.221.75/32", "D"},
+	})
+
+	// Delete left node while a right node is present
+	tree = NewTreeV4()
+	tree.Add(ipA, "A", nil)
+	tree.Add(ipB, "B", nil)
+	tree.Add(ipC, "C", nil)
+	tree.Add(ipD, "D", nil)
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "B"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.0/23", "A"},
+		{"203.143.221.75/32", "D"},
+	})
+
+	// Delete left node with a child
+	tree = NewTreeV4()
+	tree.Add(ipA, "A", nil)
+	tree.Add(ipB, "B", nil)
+	tree.Add(ipC, "C", nil)
+	tree.Add(ipD, "D", nil)
+	tree.Add(ipE, "E", nil)
+	iter = tree.Iterate()
+	for iter.Next() {
+		iter.Delete(func(payload, val GeneratedType) bool {
+			return payload == "B"
+		}, "")
+	}
+	compare([][]string{
+		{"203.143.0.0/16", "C"},
+		{"203.143.220.0/23", "A"},
+		{"203.143.220.198/32", "E"},
+		{"203.143.221.75/32", "D"},
+	})
+}
